/*
 * Ahoi. climit.c Limiter pocz. Prosze o opinie na temat tego moduu.
 * Odrobina kodu wzita z rde kernela. Dzikuje NEO2k za udostpnienie
 * maszyny do testowania i rielowi za pomoc. Miej zabawy.
 * 
 * Kompilacja: make
 * 
 * Instalacja: make load
 * 
 * Deinstalacja: make unload
 * 
 * Wszelkie opinie mile widzane. Aktualnie modu obejmuje wszystkie serwery
 * z pliku all_servers. Defaultowo limity ustawione sa na 2/per uytkownik.
 * Pozmieniaj wedug wasnego uznania MAXCONN i BGID.
 * W wypadku bdow pisz.
 * 
 * airot@toya.net.pl.
 * 
 */

/*
 * Ahoi. climit.c connect() limiter. Please send me your opinions about this
 * module. Some code riped from kernel source. Thanks goes to NEO2k for
 * testing machine, and for riel for help. ;-)
 * 
 * Compile: type 'make' Installation: type 'make load' or insmod climit.o
 * Deinstalation: type 'make unload' or rmmod climit
 * 
 * Currently the module take control on servers from file all_servers. I`m sure
 * it will change, it`s just v1.1. The limits are set to 2 connect()`s. You
 * can change it in gen.h file. Same with bad gid. In case of any
 * bugs/questions/feutures email me
 * 
 * airot@toya.net.pl
 */

/*
 * connect limiter module This program is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 675
 * Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: climit.c,v 1.1 2002/09/10 17:51:18 airot Exp $ */

#define __KERNEL__
#define MODULE

#include "gen.h"

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/modversions.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/capability.h>
#include <linux/limits.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/unistd.h>
#include <asm/unistd.h>
#include <asm/segment.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <net/sock.h>
#include <linux/version.h>


extern void    *sys_call_table[];

int             (*old_socketcall) (int call, unsigned long *args);
int             (*old_close) (int fd);
void            (*old_exit) (int status);

static u_int32_t block[MAXSERV];
static int      counter = 0;
static int      uids[MAXUIDS];

static void
first_call(void)
{
	register int    i = 0;
	lock_kernel();
	for (; i < MAXUIDS; i++)
		uids[i] = 0;
	uids[0] = -1;
#include "servers.h"		/* yea, ugly, but it`s v1.0! :) */
	unlock_kernel();
}


extern __inline__ struct socket *
socki_lookup(struct inode * inode)
{
	return &inode->u.socket_i;
}

extern __inline__ void
sockfd_put(struct socket * sock)
{
	fput(sock->file);
}


struct socket  *
sockfd_lookup(int fd)
{
	struct file    *file;
	struct inode   *inode;
	struct socket  *sock;
	if (!(file = fget(fd))) {
		return NULL;
	}
	inode = file->f_dentry->d_inode;
	if (!inode || !inode->i_sock || !(sock = socki_lookup(inode))) {
		fput(file);
		return NULL;
	}
	if (sock->file != file) {
		printk(KERN_ERR "socki_lookup: socket file changed!\n");
		sock->file = file;
	}
	return sock;
}

u_int32_t
sys_getpeername(int fd, int *err)
{
	struct socket  *sock;
	char            address[MAX_SOCK_ADDR];
	int             len;
	u_int32_t       bzium = 0;
	*err = 1;
	lock_kernel();
	if ((sock = sockfd_lookup(fd)) != NULL) {
		*err = sock->ops->getname(sock, (struct sockaddr *) address, &len, 1);
		if (!*err)
			memcpy(&bzium, address + 4, 4);
		sockfd_put(sock);
	}
	unlock_kernel();
	return bzium;
}


static void
count_em(uid_t uid)
{
	int             fd = 0, j = 0, err, i = 0;
	u_int32_t	blah = 0;
	struct files_struct *files;

	if (!(files = current->files))
		return;
	uids[uid] = 0;
	for (;;) {
		unsigned long   set;
		fd = j * __NFDBITS;
		if (fd >= files->max_fdset || fd >= files->max_fds)
			break;
		set = files->open_fds->fds_bits[j++];
		while (set) {
			if (set & 1) {
				struct file    *file = files->fd[fd];
				if (file) {
					blah = sys_getpeername(fd, &err);
					if (!err)
						for (i = 0; i < counter; i++) {
							if (block[i] == blah) 
								uids[uid]++;
						}
				}
			}
			fd++;
			set >>= 1;
		}
	}
	return;
}

static int
new_socketcall(int call, unsigned long *args)
{
	int             ret = 0;
	unsigned long   a[5];
	uid_t           uid = current->uid;
	gid_t           gid = current->gid;


	if (uids[0] != -1)
		first_call();
	if (gid != BGID || call != SYS_CONNECT)
		return old_socketcall(call, args);
	if ((ret = old_socketcall(call, args)))
		return ret;
	lock_kernel();
	if (copy_from_user(a, args, 3 * sizeof(a[0]))) {
		return -EFAULT;
	}
	count_em(uid);
	if (uids[uid] > MAXCONN) {
#ifdef DEBUG
		printk("Too many connections %d (%d)\n", uids[uid], uid);
#endif
		unlock_kernel();
		old_close(a[0]);
		return -EPERM;
	}
	unlock_kernel();
	return ret;
}

int
n_close(int fd)
{
	u_int32_t       blah;
	int             i = 0, err;
	uid_t           uid = current->uid;
	gid_t           gid = current->gid;

	if (gid != BGID)
		return old_close(fd);
	blah = sys_getpeername(fd, &err);
	if (!err)
		for (i = 0; i < counter; i++) {
			if (block[i] == blah) {
				if (uids[uid] >= 0)
					uids[uid]--;
			}
		}

	return old_close(fd);
}

void
new_exit(int status)
{
	uid_t           uid = current->uid;
	gid_t           gid = current->gid;

	if (gid != BGID)
		return old_exit(status);
	lock_kernel();
	count_em(uid);
	unlock_kernel();
	return old_exit(status);
}



int
init_module()
{
	unsigned long   flags;


	save_flags(flags);
	cli();
	old_socketcall = sys_call_table[__NR_socketcall];
	old_close = sys_call_table[__NR_close];
	old_exit = sys_call_table[__NR_exit];
	sys_call_table[__NR_exit] = new_exit;
	sys_call_table[__NR_close] = n_close;
	sys_call_table[__NR_socketcall] = new_socketcall;
	restore_flags(flags);

	return 0;
}

void
cleanup_module()
{
	unsigned long   flags;


	save_flags(flags);
	cli();
	sys_call_table[__NR_socketcall] = old_socketcall;
	sys_call_table[__NR_close] = old_close;
	sys_call_table[__NR_exit] = old_exit;
	restore_flags(flags);

}
